library(plyr)
library(pbapply)
library(MASS)
library(LaplacesDemon)
library(truncnorm)
library(stats)
library(parallel) 


############################################################
#RKDRE EM algorithm
############################################################

RKDRE_h <- function(data, formula, tol_beta = 1e-12, tol_EM = 1e-12, max.iter = 100, h){
  
  n <- nrow(data)
  
  #compute OLS estimate
  OLS_model <- lm(data = data, formula = formula)
  beta_OLS <- OLS_model$coefficients
  
  #set initial estimate
  beta <- beta_OLS
  
  #create matrix of explanatory variables with intercept
  X <- model.matrix(formula, as.data.frame(data))
  
  #column of dependent variable
  y <- data[,which(colnames(data) == all.vars(formula)[1])]
  
  #create A for in EM loop
  A <- solve( t(X) %*% X )
  
  converged_beta <- 0 
  counter <- 0
  
  while (converged_beta == 0 && counter < max.iter){ #start beta loop
    
    counter <- counter + 1
    beta_prev <- beta
    beta_EM <- beta_prev
    converged_EM <- 0
    
    #create residuals
    resid <- y - X %*% beta_prev
    
    resid_matrix <- do.call(rbind, replicate(n, t(resid), simplify=FALSE))
    
    while (converged_EM == 0){ #start EM loop
      
      beta_prev_EM <- beta_EM
      
      resid_EM <- y - X %*% beta_prev_EM
      
      p_matrix <- matrix(rep(resid_EM,n), nrow = n, ncol = n) - resid_matrix
      
      p_matrix <- exp( -0.5 * (p_matrix / h )^2 )
      
      p_matrix <- p_matrix / ( rowSums(p_matrix) )
      
      #split beta_update in two parts (beta_OLS and A %*% (B-C))
      #A is solve( t(X) %*% X) is created outside the loops
      
      PE_sum <- p_matrix %*% resid 
      
      B <- colSums(X * as.vector(PE_sum))
      
      C <- mean(PE_sum) * colSums(X)
      
      #update beta
      beta_EM <- beta_OLS - A %*% (B - C)
      
      if(sum((beta_EM - beta_prev_EM)^2) < tol_EM){
        converged_EM <- 1
        beta <- beta_EM
      }
      
    } #end of EM loop
    
    if (sum((beta - beta_prev)^2) < tol_beta){
      converged_beta <- 1
    }
  } #end of beta loop
  
  if (counter >= max.iter){
    warning('RKDRE did not converge in specified maximum number of iterations')
  }
  
  return(list(coefficients = beta, iterations = counter))
}

########################################################
#KDRE
#########################################################
KDRE_h <- function(data, formula, tol_EM = 1e-12, h){
  
  beta <- NULL
  n <- nrow(data)
  
  OLS_model <- lm(data = data, formula = formula)
  beta_OLS <- OLS_model$coefficients
  
  #create matrix of explanatory variables with intercept
  X <- model.matrix(formula, as.data.frame(data))
  
  #column of dependent variable
  y <- data[,which(colnames(data) == all.vars(formula)[1])]
  
  #create A for in EM loop
  A <- solve( t(X) %*% X )
  
  #initialize EM algorithm
  beta <- beta_OLS
  converged_EM <- 0
  
  #create residuals
  resid_OLS <- y - X %*% beta
  
  resid_matrix_OLS <- do.call(rbind, replicate(n, t(resid_OLS), simplify=FALSE))
  
  while (converged_EM == 0){ #start EM loop
    
    beta_prev <- beta
    
    resid <- y - X %*% beta_prev
    
    #create matrix of classification probabilities
    p_matrix <- matrix(rep(resid,n), nrow = n, ncol = n) - resid_matrix_OLS
    p_matrix <- exp( -0.5 * (p_matrix / h )^2  )
    p_matrix <- p_matrix / ( rowSums(p_matrix)  )
    
    PE_sum <- p_matrix %*% resid_OLS 
    
    #update beta
    beta <- beta_OLS - A %*% colSums(X * as.vector(PE_sum))
    
    if(sum((beta - beta_prev)^2) < tol_EM){
      converged_EM <- 1
    }
    
  } #end of EM loop
  
  return(list(coefficients = beta))
}

####################################################
#SBS
####################################################
SBS_h <- function(data, formula, t = 8, h){
  
  n <- nrow(data)
  
  #compute trimming parameters
  a <- t
  b <- exp(-0.5*t^2)
  c <- t
  
  OLS_model <- lm(data = data, formula = formula)
  beta <- OLS_model$coefficients
  
  #create matrix of explanatory variables with intercept
  X <- model.matrix(formula, as.data.frame(data))
  
  #column of dependent variable
  y <- data[,which(colnames(data) == all.vars(formula)[1])]
  
  #compute OLS residuals
  resid <- y - X %*% beta
  
  #create matrix for score computation
  resid_matrix <- matrix(rep(resid,n), nrow = n, ncol = n)
  
  resid_matrix <- ( resid_matrix - t(resid_matrix) ) / h
  
  exp_resid_matrix <- exp( -0.5 * (resid_matrix^2) )
  
  #compute kernel density (up to scale factor)
  f_u <- rowSums( exp_resid_matrix )
  
  #compute derivative of kernel density (up to scale factor)
  df_u <- rowSums( -1 * resid_matrix * exp_resid_matrix ) 
  
  #compute score function
  score <- 1/h * df_u / f_u
  
  #trim score
  trim_matrix <- cbind(resid,f_u,score)
  
  trimmed_score <- apply(trim_matrix, 1, function(x){
    if (abs(x[1]) < a && x[2] > b && abs(x[3]) < c ){
      x <- x[3]
    }
    else{
      x <- 0
    }
    return(x)
  })
  
  score_X <- X * trimmed_score
  
  #compute the SBS estimator
  beta <- beta - n / sum(trimmed_score^2) * solve(t(X) %*% X) %*% colSums(score_X)
  
  resid <- y - X %*% beta
  
  return(list(coefficients = beta, residuals = resid, h = h))
}

######################################################
#YDG
######################################################

YDG_h <- function(data, formula, tol_EM = 1e-9, optim = "ML", h){
  
  beta <- NULL
  n <- nrow(data)
  
  OLS_model <- lm(data = data, formula = formula)
  beta_OLS <- OLS_model$coefficients
  p <- length(beta_OLS)
  
  #create matrix of explanatory variables with intercept
  X <- model.matrix(formula, as.data.frame(data))
  
  #column of dependent variable
  y <- data[,which(colnames(data) == all.vars(formula)[1])]
  
  if (optim == "EM"){
    
    if (p == 2){
      
      #create matrix of differences in explanatory variables (xi- xj)
      X_diff <- rep(X[,2],n) - do.call(rbind, replicate(n, t(X[,2]), simplify=FALSE))
      
      #create matrix of differences in dependent variable (yi- yj)
      y_diff <- rep(y,n) - do.call(rbind, replicate(n, t(y), simplify=FALSE))
      
      beta_EM <- beta_OLS[-1]
      converged_EM <- 0
      
      while (converged_EM == 0){
        
        beta_prev_EM <- beta_EM
        
        #create residual matrix
        resid_matrix <- y_diff - X_diff * beta_prev_EM
        
        #create matrix of classification probabilities
        p_matrix <- exp(-0.5 * ( resid_matrix / h )^2 )
        #use leave one out
        diag(p_matrix) <- 0
        #divide by sum of the rows such that probabilities sum to 1
        p_matrix <- p_matrix / rowSums(p_matrix)
        
        #update beta
        beta_EM <- 1/( sum(na.omit(p_matrix * X_diff^2 ))) * sum(na.omit(p_matrix * X_diff * y_diff))  
        
        if(max((beta_EM - beta_prev_EM)^2) < tol_EM){
          converged_EM <- 1
          beta <- beta_EM
        }
      } # end of EM loop
    }
    
    if (p > 2){
      #create three-dimensional array of differences in explanatory variables (xi- xj)
      X_diff <- array(0,dim = c(n,n,(p-1)))
      for (i in 1:(p-1)){
        X_diff[,,i] <- rep(X[,(i+1)],n) - do.call(rbind, replicate(n, t(X[,(i+1)]), simplify=FALSE))
      }
      
      #create matrix of differences in dependent variable (yi- yj)
      y_diff <- rep(y,n) - do.call(rbind, replicate(n, t(y), simplify=FALSE))
      
      beta_EM <- beta_OLS[-1]
      converged_EM <- 0
      
      while (converged_EM == 0){ #start EM loop
        
        beta_prev_EM <- beta_EM
        resid <- y - as.matrix(X[,-1]) %*% beta_prev_EM
        
        #create residual matrix
        resid_matrix <- matrix(rep(resid,n), nrow = n, ncol = n)
        resid_matrix <- resid_matrix -t(resid_matrix)
        
        #create matrix of classification probabilities
        p_matrix <- exp(-0.5 * ( resid_matrix / h )^2 )
        #use leave one out
        diag(p_matrix) <- 0
        #divide by sum of the rows such that probabilities sum to 1
        p_matrix <- p_matrix / rowSums(p_matrix)
        
        #the inner apply computes the outer product X_diff[i,j,] %*% t(X_diff[i,j,])
        #the outer products are structured with array
        #sweep multiplies each outer product element with p[i,j]
        #the outer apply sums over the matrices of outer products
        A <- apply(sweep(array(apply(X_diff, c(1,2), function(x){x %*% t(x)}), dim = c(p-1,p-1,n,n)), c(3,4), FUN = "*",p_matrix), c(1,2), sum)
        B <- apply(sweep(X_diff, c(1,2), FUN = "*", (y_diff * p_matrix)),3,sum)
        
        #update beta
        beta_EM <- solve(A) %*% B
        
        if(max((beta_EM - beta_prev_EM)^2) < tol_EM){
          converged_EM <- 1
          beta <- beta_EM
        }
      } #end of EM loop
    }
  }
  
  if (optim == "ML"){
    
      fn <- function(beta, y, X, n, h){
      
      #create residuals
      resid <- y - as.matrix(X[,-1]) %*% beta
      
      #create matrix of differences
      resid_matrix <- matrix(rep(resid,n), nrow = n, ncol = n)
      resid_matrix <- ( resid_matrix - t(resid_matrix) ) / h
      
      #compute Gaussian kernel scores
      K_matrix <- exp(-0.5 * resid_matrix^2)
      #use leave one out density
      diag(K_matrix) <- 0
      
      #compute leave one out density estimates
      f_vector <- 1/( (n-1) * h ) * rowSums(K_matrix) 
      
      #compute (negative) likelihood
      fn <- -1*sum(log(f_vector))
      fn
      }
      
      #if initial value is infinite, return NULL
      if ( is.infinite(fn(beta = beta_OLS[-1],y = y, X = X, n = n, h = h)) == TRUE ){
        return(list(coefficients = rep(NaN,p)))
        warning('initial estimated log-likelihood in YDG is (minus) infinite')
      }
      
    
    beta <- try(optim(par = beta_OLS[-1], fn = fn, method = "BFGS", y = y, X = X, n = n, h = h)$par, silent = TRUE)
  }
  
  #compute intercept term
  intercept <- mean(y - as.matrix(X[,-1]) %*% beta)
  beta <- c(intercept,beta)
  
  return(list(coefficients = beta))
}

#function to perform a simulation
sim_h <- function(nrep, n = 100, nvar = 1, coef = rep(1,(nvar+1)),  
                distribution = c("normal","uniform","heavy_tail","contam_normal","multi_modal", "skewed", 
                                 "heterosked_normal","unif_normal", "log_normal","gamma","truncnorm","beta"),
                tol_beta = 1e-12, tol_EM = 1e-12, J = 3, t = 8){
  
  set.seed(1234)
  
  #######################################
  #run pbreplicate in parallel for speed
  #######################################
  
  n_cores <- detectCores() #specify number of nodes to use
  cl <- makeCluster(n_cores) #make cluster
  #functions and variables needed in cluster
  var_list <- list("get.data","KDRE", "RKDRE", "SBS", "YDG", "KDRE_h","RKDRE_h","SBS_h","YDG_h","nrep", "n",
                   "nvar", "coef", "distribution","tol_beta","tol_EM","J","t") 
  
  clusterEvalQ(cl = cl, { #libraries needed in cluster
    library(plyr)
    library(pbapply)
    library(MASS)
    library(LaplacesDemon)
    library(truncnorm)
    library(stats)
  })
  clusterExport(cl = cl, var_list, envir=environment()) #specify environment as local environment of function
  
  results <- pbreplicate(nrep, {
    
    d <- get.data(n = n, nvar = nvar, coef = coef, distribution = distribution)           
    frmla <- as.formula('y~.')
    fit.lm <- lm(formula = frmla, data = d)
    
    fit.KDRE_0.1 <- KDRE_h(data = d, formula = frmla, tol_EM = tol_EM, h = 0.1)
    fit.KDRE_0.2 <- KDRE_h(data = d, formula = frmla, tol_EM = tol_EM, h = 0.2)
    fit.KDRE_0.3 <- KDRE_h(data = d, formula = frmla, tol_EM = tol_EM, h = 0.3)
    fit.KDRE_0.4 <- KDRE_h(data = d, formula = frmla, tol_EM = tol_EM, h = 0.4)
    fit.KDRE_0.5 <- KDRE_h(data = d, formula = frmla, tol_EM = tol_EM, h = 0.5)
    fit.KDRE_0.6 <- KDRE_h(data = d, formula = frmla, tol_EM = tol_EM, h = 0.6)
    fit.KDRE_0.7 <- KDRE_h(data = d, formula = frmla, tol_EM = tol_EM, h = 0.7)
    fit.KDRE_0.8 <- KDRE_h(data = d, formula = frmla, tol_EM = tol_EM, h = 0.8)
    fit.KDRE_0.9 <- KDRE_h(data = d, formula = frmla, tol_EM = tol_EM, h = 0.9)
    fit.KDRE_norm <- KDRE(data = d, formula = frmla, tol_EM = tol_EM, bandwidth = "normal")
    fit.KDRE_rob1 <- KDRE(data = d, formula = frmla, tol_EM = tol_EM, bandwidth = "robust1")
    fit.KDRE_rob2 <- KDRE(data = d, formula = frmla, tol_EM = tol_EM, bandwidth = "robust2")
    
    fit.RKDRE_0.1 <- RKDRE_h(data = d, formula = frmla, tol_EM = tol_EM, h = 0.1)
    fit.RKDRE_0.2 <- RKDRE_h(data = d, formula = frmla, tol_EM = tol_EM, h = 0.2)
    fit.RKDRE_0.3 <- RKDRE_h(data = d, formula = frmla, tol_EM = tol_EM, h = 0.3)
    fit.RKDRE_0.4 <- RKDRE_h(data = d, formula = frmla, tol_EM = tol_EM, h = 0.4)
    fit.RKDRE_0.5 <- RKDRE_h(data = d, formula = frmla, tol_EM = tol_EM, h = 0.5)
    fit.RKDRE_0.6 <- RKDRE_h(data = d, formula = frmla, tol_EM = tol_EM, h = 0.6)
    fit.RKDRE_0.7 <- RKDRE_h(data = d, formula = frmla, tol_EM = tol_EM, h = 0.7)
    fit.RKDRE_0.8 <- RKDRE_h(data = d, formula = frmla, tol_EM = tol_EM, h = 0.8)
    fit.RKDRE_0.9 <- RKDRE_h(data = d, formula = frmla, tol_EM = tol_EM, h = 0.9)
    fit.RKDRE_norm <- RKDRE(data = d, formula = frmla, tol_EM = tol_EM, bandwidth = "normal", update_bw = FALSE)
    fit.RKDRE_rob1 <- RKDRE(data = d, formula = frmla, tol_EM = tol_EM, bandwidth = "robust1", update_bw = FALSE)
    fit.RKDRE_rob2 <- RKDRE(data = d, formula = frmla, tol_EM = tol_EM, bandwidth = "robust2", update_bw = FALSE)
    fit.RKDRE_norm_update <- RKDRE(data = d, formula = frmla, tol_EM = tol_EM, bandwidth = "normal", update_bw = TRUE)
    fit.RKDRE_rob1_update <- RKDRE(data = d, formula = frmla, tol_EM = tol_EM, bandwidth = "robust1", update_bw = TRUE)
    fit.RKDRE_rob2_update <- RKDRE(data = d, formula = frmla, tol_EM = tol_EM, bandwidth = "robust2", update_bw = TRUE)
    
    fit.SBS_0.1 <- SBS_h(data = d, formula = frmla, h = 0.1)
    fit.SBS_0.2 <- SBS_h(data = d, formula = frmla, h = 0.2)
    fit.SBS_0.3 <- SBS_h(data = d, formula = frmla, h = 0.3)
    fit.SBS_0.4 <- SBS_h(data = d, formula = frmla, h = 0.4)
    fit.SBS_0.5 <- SBS_h(data = d, formula = frmla, h = 0.5)
    fit.SBS_0.6 <- SBS_h(data = d, formula = frmla, h = 0.6)
    fit.SBS_0.7 <- SBS_h(data = d, formula = frmla, h = 0.7)
    fit.SBS_0.8 <- SBS_h(data = d, formula = frmla, h = 0.8)
    fit.SBS_0.9 <- SBS_h(data = d, formula = frmla, h = 0.9)
    fit.SBS_norm <- SBS(data = d, formula = frmla, bandwidth = "normal")
    fit.SBS_rob1 <- SBS(data = d, formula = frmla, bandwidth = "robust1")
    fit.SBS_rob2 <- SBS(data = d, formula = frmla, bandwidth = "robust2")
    
    fit.YDG_0.1 <- YDG_h(data = d, formula = frmla, optim = "ML", h = 0.1)
    fit.YDG_0.2 <- YDG_h(data = d, formula = frmla, optim = "ML", h = 0.2)
    fit.YDG_0.3 <- YDG_h(data = d, formula = frmla, optim = "ML", h = 0.3)
    fit.YDG_0.4 <- YDG_h(data = d, formula = frmla, optim = "ML", h = 0.4)
    fit.YDG_0.5 <- YDG_h(data = d, formula = frmla, optim = "ML", h = 0.5)
    fit.YDG_0.6 <- YDG_h(data = d, formula = frmla, optim = "ML", h = 0.6)
    fit.YDG_0.7 <- YDG_h(data = d, formula = frmla, optim = "ML", h = 0.7)
    fit.YDG_0.8 <- YDG_h(data = d, formula = frmla, optim = "ML", h = 0.8)
    fit.YDG_0.9 <- YDG_h(data = d, formula = frmla, optim = "ML", h = 0.9)
    fit.YDG_norm <- YDG(data = d, formula = frmla, optim = "ML", bandwidth = "normal")
    fit.YDG_rob1 <- YDG(data = d, formula = frmla, optim = "ML", bandwidth = "robust1")
    fit.YDG_rob2 <- YDG(data = d, formula = frmla, optim = "ML", bandwidth = "robust2")
    
    c(OLS = fit.lm$coefficients, KDRE_0.1 = fit.KDRE_0.1$coefficients, KDRE_0.2 = fit.KDRE_0.2$coefficients,
      KDRE_0.3 = fit.KDRE_0.3$coefficients, KDRE_0.4 = fit.KDRE_0.4$coefficients, KDRE_0.5 = fit.KDRE_0.5$coefficients,
      KDRE_0.6 = fit.KDRE_0.6$coefficients, KDRE_0.7 = fit.KDRE_0.7$coefficients, KDRE_0.8 = fit.KDRE_0.8$coefficients,
      KDRE_0.9 = fit.KDRE_0.9$coefficients, KDRE_norm = fit.KDRE_norm$coefficients, KDRE_rob1 = fit.KDRE_rob1$coefficients,
      KDRE_rob2 = fit.KDRE_rob2$coefficients, 
      RKDRE_0.1 = fit.RKDRE_0.1$coefficients, RKDRE_0.2 = fit.RKDRE_0.2$coefficients,
      RKDRE_0.3 = fit.RKDRE_0.3$coefficients, RKDRE_0.4 = fit.RKDRE_0.4$coefficients, RKDRE_0.5 = fit.RKDRE_0.5$coefficients,
      RKDRE_0.6 = fit.RKDRE_0.6$coefficients, RKDRE_0.7 = fit.RKDRE_0.7$coefficients, RKDRE_0.8 = fit.RKDRE_0.8$coefficients,
      RKDRE_0.9 = fit.RKDRE_0.9$coefficients, RKDRE_norm = fit.RKDRE_norm$coefficients, RKDRE_rob1 = fit.RKDRE_rob1$coefficients,
      RKDRE_rob2 = fit.RKDRE_rob2$coefficients, RKDRE_norm_upd = fit.RKDRE_norm_update$coefficients, RKDRE_rob1_upd = fit.RKDRE_rob1_update$coefficients,
      RKDRE_rob2_upd = fit.RKDRE_rob2_update$coefficients,
      SBS_0.1 = fit.SBS_0.1$coefficients, SBS_0.2 = fit.SBS_0.2$coefficients,
      SBS_0.3 = fit.SBS_0.3$coefficients, SBS_0.4 = fit.SBS_0.4$coefficients, SBS_0.5 = fit.SBS_0.5$coefficients,
      SBS_0.6 = fit.SBS_0.6$coefficients, SBS_0.7 = fit.SBS_0.7$coefficients, SBS_0.8 = fit.SBS_0.8$coefficients,
      SBS_0.9 = fit.SBS_0.9$coefficients, SBS_norm = fit.SBS_norm$coefficients, SBS_rob1 = fit.SBS_rob1$coefficients,
      SBS_rob2 = fit.SBS_rob2$coefficients, 
      YDG_0.1 = fit.YDG_0.1$coefficients, YDG_0.2 = fit.YDG_0.2$coefficients,
      YDG_0.3 = fit.YDG_0.3$coefficients, YDG_0.4 = fit.YDG_0.4$coefficients, YDG_0.5 = fit.YDG_0.5$coefficients,
      YDG_0.6 = fit.YDG_0.6$coefficients, YDG_0.7 = fit.YDG_0.7$coefficients, YDG_0.8 = fit.YDG_0.8$coefficients,
      YDG_0.9 = fit.YDG_0.9$coefficients, YDG_norm = fit.YDG_norm$coefficients, YDG_rob1 = fit.YDG_rob1$coefficients,
      YDG_rob2 = fit.YDG_rob2$coefficients, h_norm = 1.06*sd(fit.lm$residuals)*n^(-0.2),
      h_rob1 = 0.79 * IQR(fit.lm$residuals) * n^(-0.2), h_rob2 = 0.9 * min(sd(fit.lm$residuals),IQR(fit.lm$residuals)/1.34) * n^(-0.2))  
  }, cl = cl)
  
  stopCluster(cl) 
  
  #######################################
  #stop cluster
  #######################################
  
  results <- t(results)
  k <- ncol(results)
  #number of estimators
  q <- (k-3)/(nvar+1)
  
  #compute mean and variance for all estimates
  res <- rbind('mean' = apply(results[,c(1:(k-3))],2,function(x){mean(na.omit(x))}), 'var' = apply(results[,c(1:(k-3))],2,function(x){var(na.omit(x))})) 
  res <- rbind(res, 'RMSE' = ( sqrt((res[1,] - coef )^2 + res[2,]) ) )
  h <- rbind('mean' = apply(results[,c((k-2):k)],2,mean), 'var'= apply(results[,c((k-2):k)],2,var), 'min' = apply(results[,c((k-2):k)],2,min), 'max' = apply(results[,c((k-2):k)],2,max))
  perc_errors <- apply(results[,c(1:(k-3))],2,function(x){mean(is.nan(x))})
  
  #this is solely to make the results more conveniently comparable
  avg_summary <- matrix(0,nrow = 3, ncol = 2*q)
  for (i in 1:q){
    #intercept results
    avg_summary[1,(2*i-1)] <- abs(res[1,(i - 1 ) * (nvar + 1) + 1] - coef[1]) #bias
    avg_summary[2,(2*i-1)] <- res[2,(i - 1 ) * (nvar + 1) + 1] #variance
    avg_summary[3,(2*i-1)] <- res[3,(i - 1 ) * (nvar + 1) + 1] #RMSE
    #averaged slope results
    avg_summary[1,2*i] <- mean(abs(res[1,c(((i - 1 ) * (nvar + 1) + 2):(i *(nvar + 1)))] - coef[-1])) #mean absolute bias
    avg_summary[2,2*i] <- mean(res[2,c(((i - 1 ) * (nvar + 1) + 2):(i *(nvar + 1)))]) #mean variance
    avg_summary[3,2*i] <- sqrt( mean(res[3,c(((i - 1 ) * (nvar + 1) + 2):(i *(nvar + 1)))]^2) ) #mean RMSE
  }
  avg_summary <- round(avg_summary,digits = 4)
  rownames(avg_summary) <- c("bias","var","RMSE")
  colnames(avg_summary) <- rep(c("Intercept","Slope"),q)
  
  OLS <- avg_summary[,1:2]
  KDRE_0.1 <- avg_summary[,3:4]
  KDRE_0.2 <- avg_summary[,5:6]
  KDRE_0.3 <- avg_summary[,7:8]
  KDRE_0.4 <- avg_summary[,9:10]
  KDRE_0.5 <- avg_summary[,11:12]
  KDRE_0.6 <- avg_summary[,13:14]
  KDRE_0.7 <- avg_summary[,15:16]
  KDRE_0.8 <- avg_summary[,17:18]
  KDRE_0.9 <- avg_summary[,19:20]
  KDRE_norm <- avg_summary[,21:22]
  KDRE_rob1 <- avg_summary[,23:24]
  KDRE_rob2 <- avg_summary[,25:26]
  
  RKDRE_0.1 <- avg_summary[,27:28]
  RKDRE_0.2 <- avg_summary[,29:30]
  RKDRE_0.3 <- avg_summary[,31:32]
  RKDRE_0.4 <- avg_summary[,33:34]
  RKDRE_0.5 <- avg_summary[,35:36]
  RKDRE_0.6 <- avg_summary[,37:38]
  RKDRE_0.7 <- avg_summary[,39:40]
  RKDRE_0.8 <- avg_summary[,41:42]
  RKDRE_0.9 <- avg_summary[,43:44]
  RKDRE_norm <- avg_summary[,45:46]
  RKDRE_rob1 <- avg_summary[,47:48]
  RKDRE_rob2 <- avg_summary[,49:50]
  RKDRE_norm_update <- avg_summary[,51:52]
  RKDRE_rob1_update <- avg_summary[,53:54]
  RKDRE_rob2_update <- avg_summary[,55:56]
  
  SBS_0.1 <- avg_summary[,57:58]
  SBS_0.2 <- avg_summary[,59:60]
  SBS_0.3 <- avg_summary[,61:62]
  SBS_0.4 <- avg_summary[,63:64]
  SBS_0.5 <- avg_summary[,65:66]
  SBS_0.6 <- avg_summary[,67:68]
  SBS_0.7 <- avg_summary[,69:70]
  SBS_0.8 <- avg_summary[,71:72]
  SBS_0.9 <- avg_summary[,73:74]
  SBS_norm <- avg_summary[,75:76]
  SBS_rob1 <- avg_summary[,77:78]
  SBS_rob2 <- avg_summary[,79:80]
  
  YDG_0.1 <- avg_summary[,81:82]
  YDG_0.2 <- avg_summary[,83:84]
  YDG_0.3 <- avg_summary[,85:86]
  YDG_0.4 <- avg_summary[,87:88]
  YDG_0.5 <- avg_summary[,89:90]
  YDG_0.6 <- avg_summary[,91:92]
  YDG_0.7 <- avg_summary[,93:94]
  YDG_0.8 <- avg_summary[,95:96]
  YDG_0.9 <- avg_summary[,97:98]
  YDG_norm <- avg_summary[,99:100]
  YDG_rob1 <- avg_summary[,101:102]
  YDG_rob2 <- avg_summary[,103:104]
  
  return(list(summary = list(OLS = OLS,
            KDRE = list(KDRE_0.1 = KDRE_0.1, KDRE_0.2 = KDRE_0.2, KDRE_0.3 = KDRE_0.3, KDRE_0.4 = KDRE_0.4, 
                        KDRE_0.5 = KDRE_0.5, KDRE_0.6 = KDRE_0.6, KDRE_0.7 = KDRE_0.7, KDRE_0.8 = KDRE_0.8, 
                        KDRE_0.9 = KDRE_0.9, KDRE_norm = KDRE_norm, KDRE_rob1 = KDRE_rob1, KDRE_rob2 = KDRE_rob2),
            RKDRE = list(RKDRE_0.1 = RKDRE_0.1, RKDRE_0.2 = RKDRE_0.2, RKDRE_0.3 = RKDRE_0.3, RKDRE_0.4 = RKDRE_0.4, 
                         RKDRE_0.5 = RKDRE_0.5, RKDRE_0.6 = RKDRE_0.6, RKDRE_0.7 = RKDRE_0.7, RKDRE_0.8 = RKDRE_0.8, 
                         RKDRE_0.9 = RKDRE_0.9, RKDRE_norm = RKDRE_norm, RKDRE_rob1 = RKDRE_rob1, RKDRE_rob2 = RKDRE_rob2,
                         RKDRE_norm_update = RKDRE_norm_update, RKDRE_rob1_update = RKDRE_rob1_update, 
                         RKDRE_rob2_update = RKDRE_rob2_update),
            SBS = list(SBS_0.1 = SBS_0.1, SBS_0.2 = SBS_0.2, SBS_0.3 = SBS_0.3, SBS_0.4 = SBS_0.4, 
                        SBS_0.5 = SBS_0.5, SBS_0.6 = SBS_0.6, SBS_0.7 = SBS_0.7, SBS_0.8 = SBS_0.8, 
                        SBS_0.9 = SBS_0.9, SBS_norm = SBS_norm, SBS_rob1 = SBS_rob1, SBS_rob2 = SBS_rob2),
            YDG = list(YDG_0.1 = YDG_0.1, YDG_0.2 = YDG_0.2, YDG_0.3 = YDG_0.3, YDG_0.4 = YDG_0.4, 
                        YDG_0.5 = YDG_0.5, YDG_0.6 = YDG_0.6, YDG_0.7 = YDG_0.7, YDG_0.8 = YDG_0.8, 
                        YDG_0.9 = YDG_0.9, YDG_norm = YDG_norm, YDG_rob1 = YDG_rob1, YDG_rob2 = YDG_rob2)),
            results = res, bandwidth = h, estimates = results, perc_errors_YDG = perc_errors))
}


####################################
#Table bandwidth normal
normal.bw.res.n100 <- sim_h(nrep = 500, n = 100, nvar = 1, coef = c(-1,1), distribution = "normal")
save(normal.bw.res.n100, file = "normal.bw.res.n100.RData")

contam_normal.bw.res.n100 <- sim_h(nrep = 500, n = 100, nvar = 1, coef = c(-1,1), distribution = "contam_normal")
save(contam_normal.bw.res.n100, file = "contam_normal.bw.res.n100.RData")

heavy_tail.bw.res.n100 <- sim_h(nrep = 500, n = 100, nvar = 1, coef = c(-1,1), distribution = "heavy_tail")
save(heavy_tail.bw.res.n100, file = "heavy_tail.bw.res.n100.RData")

multi_modal.bw.res.n100 <- sim_h(nrep = 500, n = 100, nvar = 1, coef = c(-1,1), distribution = "multi_modal")
save(multi_modal.bw.res.n100, file = "multi_modal.bw.res.n100.276.RData")

uniform.bw.res.n100 <- sim_h(nrep = 500, n = 100, nvar = 1, coef = c(-1,1), distribution = "uniform")
save(uniform.bw.res.n100, file = "uniform.bw.res.n100.RData")

beta.bw.res.n100 <- sim_h(nrep = 500, n = 100, nvar = 1, coef = c(-1,1), distribution = "beta")
save(beta.bw.res.n100, file = "beta.bw.res.n100.RData")

skewed.bw.res.n100 <- sim_h(nrep = 500, n = 100, nvar = 1, coef = c(-1,1), distribution = "skewed")
save(skewed.bw.res.n100, file = "skewed.bw.res.n100.RData")

log_normal.bw.res.n100 <- sim_h(nrep = 500, n = 100, nvar = 1, coef = c(-1,1), distribution = "log_normal")
save(log_normal.bw.res.n100, file = "log_normal.bw.res.n100.RData")

gamma.bw.res.res.n100 <- sim_h(nrep = 500, n = 100, nvar = 1, coef = c(-1,1), distribution = "gamma")

#################################################

