#install.packages("plyr")
#install.packages("pbapply")
#install.packages("MASS")
#install.packages("LaplacesDemon")
#install.packages("truncnorm")
#install.packages("stats")
#install.packages("parallel")
#install.packages("quantreg")
#install.packages("sn")
#install.packages("L1pack")

library(plyr)
library(pbapply)
library(MASS)
library(LaplacesDemon)
library(truncnorm)
library(stats)
library(parallel) 
library(quantreg)
library(sn)
library(L1pack)
library(robustbase)

############################################################
#RKDRE EM algorithm
############################################################

RKDRE <- function(data, formula, tol_beta = 1e-12, tol_EM = 1e-12, max.iter = 100, 
                  bandwidth = c("normal","robust1","robust2","SJ","bcv","ucv"), update_bw = FALSE, std_error = FALSE, r = 0.25, a = 3, 
                  h.SJ = FALSE){
  
  n <- nrow(data)
  resid <- NULL
  h <- NULL
 
  #compute OLS estimate
  OLS_model <- lm(formula = formula, data = data)
  beta_OLS <- OLS_model$coefficients
  
  #set initial estimate
  beta <- beta_OLS
  
  #set initial bandwidth
  if (bandwidth == "normal"){
    h <- 1.06 * sd(OLS_model$residuals) * n^(-0.2)
  } else if (bandwidth == "robust1"){
    h <- 0.79 * IQR(OLS_model$residuals) * n^(-0.2)
  } else if (bandwidth == "robust2"){
    h <- 0.9 * min(sd(OLS_model$residuals),IQR(OLS_model$residuals)/1.34) * n^(-0.2)
  } else if (bandwidth == "SJ"){
    h <- bw.SJ(OLS_model$residuals)
  } else if (bandwidth == "bcv"){
    h <- bw.bcv(OLS_model$residuals)
  } else if (bandwidth == "ucv"){
    h <- bw.ucv(OLS_model$residuals)
  }
  
  #create matrix of explanatory variables with intercept
  X <- model.matrix(formula, as.data.frame(data))
  
  #column of dependent variable
  y <- data[,which(colnames(data) == all.vars(formula)[1])]
  
  #create A for in EM loop
  A <- solve( t(X) %*% X )
  
  converged_beta <- 0 
  counter <- 0
  
  while (converged_beta == 0 && counter < max.iter){ #start beta loop
    
    counter <- counter + 1
    beta_prev <- beta
    beta_EM <- beta_prev
    converged_EM <- 0
    
    #create residuals
    resid <- y - X %*% beta_prev
    
    if (update_bw == TRUE){
      #update bandwidth
      if (bandwidth == "normal"){
        h <- 1.06 * sd(resid) * n^(-0.2)
      } else if (bandwidth == "robust1"){
        h <- 0.79 * IQR(resid) * n^(-0.2)
      } else if (bandwidth == "robust2"){
        h <- 0.9 * min(sd(resid),IQR(resid)/1.34) * n^(-0.2)
      } else if (bandwidth == "SJ"){
        h <- bw.SJ(resid)
      } else if (bandwidth == "bcv"){
        h <- bw.bcv(resid)
      } else if (bandwidth == "ucv"){
        h <- bw.ucv(resid)
      }
    }
    
    resid_matrix <- do.call(rbind, replicate(n, t(resid), simplify=FALSE))
    
    while (converged_EM == 0){ #start EM loop
      
      beta_prev_EM <- beta_EM
      resid_EM <- y - X %*% beta_prev_EM
      
      p_matrix <- matrix(rep(resid_EM,n), nrow = n, ncol = n) - resid_matrix
      
      p_matrix <- exp( -0.5 * (p_matrix / h )^2 )
      
      p_matrix <- p_matrix / ( rowSums(p_matrix) )
      
      #split beta_update in two parts (beta_OLS and A %*% (B-C))
      #A is solve( t(X) %*% X) is created outside the loops
      
      PE_sum <- p_matrix %*% resid 
      
      B <- colSums(X * as.vector(PE_sum))
      
      C <- mean(PE_sum) * colSums(X)
      
      #update beta
      beta_EM <- beta_OLS - A %*% (B - C)
      
      
      if(max((beta_EM - beta_prev_EM)^2) < tol_EM){
        converged_EM <- 1
        beta <- beta_EM
      }
      
    } #end of EM loop
    
    if (max((beta - beta_prev)^2) < tol_beta){
      converged_beta <- 1
      resid <- y - X %*% beta
    }
  } #end of beta loop
  
  if (counter >= max.iter){
    warning('RKDRE did not converge in specified maximum number of iterations')
  }
  
  return(list(coefficients = beta, iterations = counter, residuals = resid))
}

############################################################
#KDRE EM algorithm desribed in (Yao & Zhao, 2013)
############################################################

KDRE <- function(data, formula, tol_EM = 1e-12, bandwidth = c("normal","robust1","robust2","SJ","bcv","ucv")){
  
  beta <- NULL
  h <- NULL
  resid <- NULL
  n <- nrow(data)
  
  OLS_model <- lm(data = data, formula = formula)
  beta_OLS <- OLS_model$coefficients
  
  
  #set initial bandwidth
  if (bandwidth == "normal"){
    h <- 1.06 * sd(OLS_model$residuals) * n^(-0.2)
  } else if (bandwidth == "robust1"){
    h <- 0.79 * IQR(OLS_model$residuals) * n^(-0.2)
  } else if (bandwidth == "robust2"){
    h <- 0.9 * min(sd(OLS_model$residuals),IQR(OLS_model$residuals)/1.34) * n^(-0.2)
  } else if (bandwidth == "SJ"){
    h <- bw.SJ(OLS_model$residuals)
  } else if (bandwidth == "bcv"){
    h <- bw.bcv(OLS_model$residuals)
  } else if (bandwidth == "ucv"){
    h <- bw.ucv(OLS_model$residuals)
  }
  
  #create matrix of explanatory variables with intercept
  X <- model.matrix(formula, as.data.frame(data))
  
  #column of dependent variable
  y <- data[,which(colnames(data) == all.vars(formula)[1])]
  
  #create A for in EM loop
  A <- solve( t(X) %*% X )
  
  #initialize EM algorithm
  beta <- beta_OLS
  converged_EM <- 0
  
  #create residuals
  resid_OLS <- y - X %*% beta
  
  resid_matrix_OLS <- do.call(rbind, replicate(n, t(resid_OLS), simplify=FALSE))
  
  while (converged_EM == 0){ #start EM loop
    
    beta_prev <- beta
    
    resid <- y - X %*% beta_prev
    
    #create matrix of classification probabilities
    p_matrix <- matrix(rep(resid,n), nrow = n, ncol = n) - resid_matrix_OLS
    p_matrix <- exp( -0.5 * (p_matrix / h )^2  )
    p_matrix <- p_matrix / ( rowSums(p_matrix)  )
    
    PE_sum <- p_matrix %*% resid_OLS 
    
    #update beta
    beta <- beta_OLS - A %*% colSums(X * as.vector(PE_sum))
    
    if(max((beta - beta_prev)^2) < tol_EM){
      converged_EM <- 1
      resid <- y - X %*% beta
    }
    
  } #end of EM loop
  
  return(list(coefficients = beta, residuals = resid))
}

############################################################
#SBS (Stone, Bickel, Schick) algorithm 
############################################################

SBS <- function(data, formula, t = 8, bandwidth = c("normal","robust1","robust2","SJ","bcv","ucv")){
  
  n <- nrow(data)
  h <- NULL
  
  #compute trimming parameters
  a <- t
  b <- exp(-0.5*t^2)
  c <- t
  
  OLS_model <- lm(data = data, formula = formula)
  beta <- OLS_model$coefficients
  
  #set initial bandwidth
  if (bandwidth == "normal"){
    h <- 1.06 * sd(OLS_model$residuals) * n^(-0.2)
  } else if (bandwidth == "robust1"){
    h <- 0.79 * IQR(OLS_model$residuals) * n^(-0.2)
  } else if (bandwidth == "robust2"){
    h <- 0.9 * min(sd(OLS_model$residuals),IQR(OLS_model$residuals)/1.34) * n^(-0.2)
  } else if (bandwidth == "SJ"){
    h <- bw.SJ(OLS_model$residuals)
  } else if (bandwidth == "bcv"){
    h <- bw.bcv(OLS_model$residuals)
  } else if (bandwidth == "ucv"){
    h <- bw.ucv(OLS_model$residuals)
  }
  
  #create matrix of explanatory variables with intercept
  X <- model.matrix(formula, as.data.frame(data))
  
  #column of dependent variable
  y <- data[,which(colnames(data) == all.vars(formula)[1])]
  
  #compute OLS residuals
  resid <- y - X %*% beta
  
  #create matrix for score computation
  resid_matrix <- matrix(rep(resid,n), nrow = n, ncol = n)
  
  resid_matrix <- ( resid_matrix - t(resid_matrix) ) / h
  
  exp_resid_matrix <- exp( -0.5 * (resid_matrix^2) )
  
  #compute kernel density (up to scale factor)
  f_u <- rowSums( exp_resid_matrix )
  
  #compute derivative of kernel density (up to scale factor)
  df_u <- rowSums( -1 * resid_matrix * exp_resid_matrix ) 
  
  #compute score function
  score <- 1/h * df_u / f_u
  
  #trim score
  trim_matrix <- cbind(resid,f_u,score)
  
  trimmed_score <- apply(trim_matrix, 1, function(x){
    if (abs(x[1]) < a && x[2] > b && abs(x[3]) < c ){
      x <- x[3]
    }
    else{
      x <- 0
    }
    return(x)
  })
  
  score_X <- X * trimmed_score
  
  #compute the SBS estimator
  beta <- beta - n / sum(trimmed_score^2) * solve(t(X) %*% X) %*% colSums(score_X)
  
  resid <- y - X %*% beta
  
  return(list(coefficients = beta, residuals = resid, h = h))
}

############################################################
#LGMM(S) algorithm described in Newey (1988)
############################################################

LGMM <- function(data, formula, J = 3, symmetric = TRUE, transformed = TRUE, weighted = FALSE){
  
  n <- nrow(data)
  resid_transformed <- NULL
  resid_weighted <- NULL
  zeta_matrix <- NULL
  w_matrix <- NULL
  
  #create matrix of explanatory variables with intercept
  X <- model.matrix(formula, as.data.frame(data))
  
  #column of dependent variable
  y <- data[,which(colnames(data) == all.vars(formula)[1])]
  
  #starting values for optimization routine
  OLS_model <- lm(data = data, formula = formula)
  beta <- OLS_model$coefficients
  
  sigma <- sd(OLS_model$residuals)
  
  #create residuals
  resid <- y - X %*% beta
  
  if (symmetric == TRUE){
    if (transformed == FALSE && weighted == FALSE){
      
      #create zeta matrix
      zeta_matrix <- resid
      for (j in 2:J){
        zeta_matrix <- cbind(zeta_matrix,resid^( 2*j -1 ))
      }
      #create w_matrix
      w_matrix <- rep(1,n)
      for (j in 2:J){
        w_matrix <- cbind(w_matrix,( 2*j-1 )*resid^( 2*j - 2 ))
      }
    }
    
    if (transformed == TRUE){
      
      resid_transformed <- resid / (1 + abs(resid))
      
      #create zeta matrix
      zeta_matrix <- resid_transformed
      for (j in 2:J){
        zeta_matrix <- cbind(zeta_matrix,resid_transformed^( 2*j -1 ))
      }
      
      zeta_matrix <- scale(zeta_matrix, center = TRUE, scale = FALSE)
      
      #create vector that is 1 (-1) for positive (negative) residuals
      ind <- rep(-1,n)
      ind[which(resid >= 0)] <- 1
      
      #compute 'inner' derivative of w
      w_deriv <- ( rep(1,n) + abs(resid) - resid * ind ) / ( 1 + abs(resid) )^2
      w_matrix <- w_deriv
      for (j in 2:J){
        w_matrix <- cbind(w_matrix,( 2*j-1 ) * resid_transformed^( 2*j - 2 ) * w_deriv)
      }
      
    }
    
    if (weighted == TRUE){
      #create weights of residuals
      resid_weights <- as.vector(exp(-0.5 * resid^2))
      
      zeta_matrix <- resid
      for (j in 2:J){
        zeta_matrix <- cbind(zeta_matrix,resid^( 2*j -1 ))
      }
      zeta_matrix <- zeta_matrix * resid_weights 
      
      zeta_matrix <- scale(zeta_matrix, center = TRUE, scale = FALSE)
      
      w_matrix <- ( c(1,n) - resid^2 ) * resid_weights
      for (j in 2:J){
        w_matrix <- cbind(w_matrix, ( (2*j-1) * resid^(2*j-2) - resid^(2*j) ) * resid_weights )
      }
    }
  }
  
  if (symmetric == FALSE){
    if (transformed == FALSE && weighted == FALSE){
      
      #create zeta matrix
      zeta_matrix <- resid
      for (j in 2:J){
        zeta_matrix <- cbind( zeta_matrix,resid^j )
      }
      zeta_matrix <- scale(zeta_matrix, center = TRUE, scale = FALSE)
      #create w_matrix
      w_matrix <- rep(1,n)
      for (j in 2:J){
        w_matrix <- cbind(w_matrix, j * resid^( j - 1 ))
      }
      
    }
    if (transformed == TRUE){
      
      resid_transformed <- resid / (1 + abs(resid))
      
      #create zeta matrix
      zeta_matrix <- resid_transformed
      for (j in 2:J){
        zeta_matrix <- cbind(zeta_matrix,resid_transformed^j)
      }
      
      zeta_matrix <- scale(zeta_matrix, center = TRUE, scale = FALSE)
      
      #create vector that is 1 (-1) for positive (negative) residuals
      ind <- rep(-1,n)
      ind[which(resid >= 0)] <- 1
      
      #compute 'inner' derivative of w
      w_deriv <- ( rep(1,n) + abs(resid) - resid * ind ) / ( 1 + abs(resid) )^2
      w_matrix <- w_deriv
      for (j in 2:J){
        w_matrix <- cbind(w_matrix, j * resid_transformed^( j-1 ) * w_deriv)
      }
    }
    if (weighted == TRUE){
      
      #create weights of residuals
      resid_weights <- as.vector(exp(-0.5 * resid^2))
      
      zeta_matrix <- resid
      for (j in 2:J){
        zeta_matrix <- cbind(zeta_matrix,resid^( 2*j -1 ))
      }
      zeta_matrix <- zeta_matrix * resid_weights 
      
      zeta_matrix <- scale(zeta_matrix, center = TRUE, scale = FALSE)
      
      w_matrix <- ( c(1,n) - resid^2 ) * resid_weights
      for (j in 2:J){
        w_matrix <- cbind(w_matrix, ( j * resid^(j-1) - resid^(j+1) ) * resid_weights )
      }
    }
  }
  
  #create covariance matrix of moments
  V <- ( n-1 ) / n * cov(zeta_matrix)
  
  #compute sample moments
  w <- colMeans(w_matrix)
  
  #precompute some matrices
  A <- t(X) %*% X
  wX <- kronecker(t(w),A)
  VinvAinv <- kronecker(solve(V),solve(A))
  
  #update beta
  beta <- beta + solve( wX %*% VinvAinv %*% t(wX) ) %*% wX %*% VinvAinv %*% kronecker(diag(J),t(X)) %*% as.vector(zeta_matrix)
  
  return(list(coefficients = beta))
} 

############################################################
#YDG algorithm described in (Yuan and De Gooijer, 2007)
############################################################

YDG <- function(data, formula, tol_EM = 1e-10, optim = c("EM", "ML"), bandwidth = c("normal","robust1","robust2","SJ","bcv","ucv")){
  
  beta <- NULL
  h <- NULL
  n <- nrow(data)
  
  OLS_model <- lm(data = data, formula = formula)
  beta_OLS <- OLS_model$coefficients
  p <- length(beta_OLS)
  
  #set initial bandwidth
  if (bandwidth == "normal"){
    h <- 1.06 * sd(OLS_model$residuals) * n^(-0.2)
  } else if (bandwidth == "robust1"){
    h <- 0.79 * IQR(OLS_model$residuals) * n^(-0.2)
  } else if (bandwidth == "robust2"){
    h <- 0.9 * min(sd(OLS_model$residuals),IQR(OLS_model$residuals)/1.34) * n^(-0.2)
  } else if (bandwidth == "SJ"){
    h <- bw.SJ(OLS_model$residuals)
  } else if (bandwidth == "bcv"){
    h <- bw.bcv(OLS_model$residuals)
  } else if (bandwidth == "ucv"){
    h <- bw.ucv(OLS_model$residuals)
  }
  
  #create matrix of explanatory variables with intercept
  X <- model.matrix(formula, as.data.frame(data))
  
  #column of dependent variable
  y <- data[,which(colnames(data) == all.vars(formula)[1])]
  
  if (optim == "EM"){
    
    if (p == 2){
      
      #create matrix of differences in explanatory variables (xi- xj)
      X_diff <- rep(X[,2],n) - do.call(rbind, replicate(n, t(X[,2]), simplify=FALSE))
      
      #create matrix of differences in dependent variable (yi- yj)
      y_diff <- rep(y,n) - do.call(rbind, replicate(n, t(y), simplify=FALSE))
     
      beta_EM <- beta_OLS[-1]
      converged_EM <- 0
      
      while (converged_EM == 0){
        
        beta_prev_EM <- beta_EM
        
        #create residual matrix
        resid_matrix <- y_diff - X_diff * beta_prev_EM
        
        #create matrix of classification probabilities
        p_matrix <- exp(-0.5 * ( resid_matrix / h )^2 )
        #use leave one out
        diag(p_matrix) <- 0
        #divide by sum of the rows such that probabilities sum to 1
        p_matrix <- p_matrix / rowSums(p_matrix)
        
        #update beta
        beta_EM <- 1/( sum(p_matrix * X_diff^2 )) * sum(p_matrix * X_diff * y_diff)  
        
        if(max((beta_EM - beta_prev_EM)^2) < tol_EM){
          converged_EM <- 1
          beta <- beta_EM
        }
      } # end of EM loop
    }
    
    if (p > 2){
      #create three-dimensional array of differences in explanatory variables (xi- xj)
      X_diff <- array(0,dim = c(n,n,(p-1)))
      for (i in 1:(p-1)){
        X_diff[,,i] <- rep(X[,(i+1)],n) - do.call(rbind, replicate(n, t(X[,(i+1)]), simplify=FALSE))
      }
      
      #create matrix of differences in dependent variable (yi- yj)
      y_diff <- rep(y,n) - do.call(rbind, replicate(n, t(y), simplify=FALSE))
      
      beta_EM <- beta_OLS[-1]
      converged_EM <- 0
      
      while (converged_EM == 0){ #start EM loop
        
        beta_prev_EM <- beta_EM
        resid <- y - as.matrix(X[,-1]) %*% beta_prev_EM
        
        #create residual matrix
        resid_matrix <- matrix(rep(resid,n), nrow = n, ncol = n)
        resid_matrix <- resid_matrix -t(resid_matrix)
        
        #create matrix of classification probabilities
        p_matrix <- exp(-0.5 * ( resid_matrix / h )^2 )
        #use leave one out
        diag(p_matrix) <- 0
        #divide by sum of the rows such that probabilities sum to 1
        p_matrix <- p_matrix / rowSums(p_matrix)
        
        #the inner apply computes the outer product X_diff[i,j,] %*% t(X_diff[i,j,])
        #the outer products are structured with array
        #sweep multiplies each outer product element with p[i,j]
        #the outer apply sums over the matrices of outer products
        A <- apply(sweep(array(apply(X_diff, c(1,2), function(x){x %*% t(x)}), dim = c(p-1,p-1,n,n)), c(3,4), FUN = "*",p_matrix), c(1,2), sum)
        B <- apply(sweep(X_diff, c(1,2), FUN = "*", (y_diff * p_matrix)),3,sum)
        
        #update beta
        beta_EM <- solve(A) %*% B
        
        if(max((beta_EM - beta_prev_EM)^2) < tol_EM){
          converged_EM <- 1
          beta <- beta_EM
        }
      } #end of EM loop
    }
  }
  
  if (optim == "ML"){
    
    fn <- function(beta, y, X, n, h){
      
      #create residuals
      resid <- y - as.matrix(X[,-1]) %*% beta
      
      #create matrix of differences
      resid_matrix <- matrix(rep(resid,n), nrow = n, ncol = n)
      resid_matrix <- ( resid_matrix - t(resid_matrix) ) / h
      
      #compute Gaussian kernel scores
      K_matrix <- exp(-0.5 * resid_matrix^2)
      #use leave one out density
      diag(K_matrix) <- 0
      
      #compute leave one out density estimates
      f_vector <- 1/( (n-1) * h ) * rowSums(K_matrix) 
      
      #compute (negative) likelihood
      fn <- -1*sum(log(f_vector))
      fn
    }
    
    #if initial value is infinite, return NULL
    if ( is.infinite(fn(beta = beta_OLS[-1],y = y, X = X, n = n, h = h)) == TRUE ){
      return(list(coefficients = rep(NaN,p)))
      warning('initial estimated log-likelihood in YDG is (minus) infinite')
    }
    
    beta <- optim(par = beta_OLS[-1], fn = fn, method = "BFGS", y = y, X = X, n = n, h = h)$par
  }
  
  #compute intercept term
  intercept <- mean(y - as.matrix(X[,-1]) %*% beta)
  beta <- c(intercept,beta)
  
  return(list(coefficients = beta))
} 

############################################################
#Simulation functions
############################################################

#function that generates data
get.data <- function(n = 100, nvar = 1, coef = rep(1,(nvar+1)), 
                     distribution = c("normal","uniform","heavy_tail", 
                    "contam_normal", "multi_modal", "skewed","heterosked_normal","unif_normal",
                    "log_normal","gamma","truncnorm","beta","cauchy")){  
  
  #create disturbances
  disturb <- NULL
  
  if (distribution == "normal"){
    disturb <- rnorm(n, 0, 1)
  }
  
  if (distribution == "uniform"){
    disturb <- runif(n, min = -0.5*sqrt(12), max = 0.5*sqrt(12))
  }
  
  if (distribution == "heavy_tail"){
    disturb <- rt(n, 2)
  }
  
  if (distribution == "contam_normal"){
    p <- rbern(n, 0.9)
    disturb <- p * rnorm(n, 0, (1/3)) + (rep(1,n)-p) * rnorm(n, 0, 3)
  }
  
  if (distribution == "multi_modal"){
    p <- rbern(n, 0.5)
    disturb <- (p * rnorm(n, -3, 1) + (rep(1,n)-p) * rnorm(n, 3, 1)) / sqrt(10)
  }
  
  if (distribution == "skewed"){
    p <- rbern(n, 0.3)
    disturb <- p * rnorm(n, -1.4, 1) + (rep(1,n)-p) * rnorm(n, 0.6, 0.4)/sqrt(1.252)
  }
  
  if (distribution == "heterosked_normal"){
    disturb <- rnorm(n, mean = 0, sd = runif(n, min = 0, max = 3))
  }
  
  if (distribution == "unif_normal"){
    p <- rbern(n, 0.5)
    disturb <- p * runif(n, min = -5, max = 5) + (rep(1,n)-p) * rnorm(n, 0, 1)
  }
  
  if (distribution == "log_normal"){
    disturb <- ( rlnorm(n, meanlog = 0, sdlog = 1) - exp(0.5) ) / sqrt( (exp(1)-1)*exp(1) )
  }
  
  if (distribution == "gamma"){
    disturb <- (rgamma(n, shape =2, rate = 0.5) - rep(4,n)) / sqrt(8)
  }
  
  if (distribution == "beta"){
    disturb <- ( rbeta(n, shape1 =2, shape2 = 2) - rep(0.5,n) ) / sqrt(0.05) 
  }
  
  if (distribution == "truncnorm"){
    p <- rbern(n, 0.5)
    disturb <- p * rtruncnorm(n, a = -10, b = 10, mean = -1, sd = 1) + (rep(1,n)-p) * rtruncnorm(n, a = -10, b = 10, mean = 1, sd = 1)
  }
  if (distribution == "cauchy"){
    disturb <- rcauchy(n)
  }
  
  #create data
  X <- replicate(nvar,rnorm(n, mean = 0, sd = 1))
  y <- as.matrix(cbind(rep(1,n),X)) %*% coef + disturb 
  
  #output
  output <- as.data.frame(cbind(y,X))
  colnames(output)[1] <- 'y'
  return(output)
}

#function to perform a simulation
sim <- function(nrep, n = 100, nvar = 1, coef = rep(1,(nvar+1)),  
                distribution = c("normal","uniform","heavy_tail","contam_normal","multi_modal", "skewed", 
                                 "heterosked_normal","unif_normal", "log_normal","gamma","truncnorm","beta","cauchy"),
                tol_beta = 1e-12, tol_EM = 1e-12, J = 3, t = 8, bandwidth = c("normal","robust1","robust2","SJ","bcv","ucv"),
                update_bw = FALSE){
  set.seed(123)
  
  #######################################
  #run pbreplicate in parallel for speed
  #######################################
  
  n_cores <- detectCores()  #specify number of nodes to use
  cl <- makeCluster(n_cores) #make cluster
  #functions and variables needed in cluster
  var_list <- list("get.data","KDRE","RKDRE","SBS","LGMM","YDG","nrep", "n",
                   "nvar", "coef", "distribution","tol_beta","tol_EM","J","t","bandwidth", "update_bw") 
                    
  clusterEvalQ(cl = cl, { #libraries needed in cluster
    library(plyr)
    library(pbapply)
    library(MASS)
    library(LaplacesDemon)
    library(truncnorm)
    library(stats)
    library(fGarch)
    library(sn)
    library(robustbase)
  })
  clusterExport(cl = cl, var_list, envir=environment()) #specify environment as local environment of function
  
  results <- pbreplicate(nrep, {
    d <- get.data(n = n, nvar = nvar, coef = coef, distribution = distribution)           
    frmla <- as.formula('y~.')
    fit.lm <- lm(formula = frmla, data = d)
    fit.KDRE <- KDRE(data = d, formula = frmla, tol_EM = tol_EM, bandwidth = "normal")
    fit.RKDRE <- RKDRE(data = d, formula = frmla, tol_beta = tol_beta, tol_EM = tol_EM, bandwidth = bandwidth, update_bw = update_bw)
    fit.SBS <- SBS(data = d, formula = frmla, t = t, bandwidth = "robust2")
    fit.LGMMS.trans <- LGMM(data = d, formula = frmla, J = J, transformed = TRUE, weighted = FALSE, symmetric = TRUE)
    fit.LGMM.trans <- LGMM(data = d, formula = frmla, J = J, transformed = TRUE, symmetric = FALSE)
    fit.YDG <-YDG(data = d, formula = frmla, optim = "ML", bandwidth = "normal")
    c(OLS = fit.lm$coefficients, KDRE = fit.KDRE$coefficients, RKDRE = fit.RKDRE$coefficients,
      SBS = fit.SBS$coefficients, LGMMS = fit.LGMMS.trans$coefficients, LGMM = fit.LGMM.trans$coefficients,
      YDG = fit.YDG$coefficients, iterations_RKDRE = fit.RKDRE$iterations)  
  }, cl = cl)
  
  stopCluster(cl) 
  
  #######################################
  #stop cluster
  #######################################
  
  results <- t(results)
  k <- ncol(results)
  #number of estimators
  q <- (k-1)/(nvar+1)
  
  #compute mean and variance for all estimates (and percentage of errors for YDG)
  res <- rbind('mean' = apply(results[,-k],2,function(x){mean(na.omit(x))}), 'var' = apply(results[,-k],2,function(x){var(na.omit(x))})) 
  res <- rbind(res, 'RMSE' = ( sqrt((res[1,] - coef )^2 + res[2,]) ) )
  perc_errors <- apply(results[,-k],2,function(x){mean(is.nan(x))})
  iter <- rbind('mean' = mean(results[,k]), 'var'= var(results[,k]), 'min' = min(results[,k]), 'max' = max(results[,k]))
  
  
  #this is solely to make the results more conveniently comparable
  avg_summary <- matrix(0,nrow = 3, ncol = 2*q)
  for (i in 1:q){
    #intercept results
    avg_summary[1,(2*i-1)] <- abs(res[1,(i - 1 ) * (nvar + 1) + 1] - coef[1]) #bias
    avg_summary[2,(2*i-1)] <- res[2,(i - 1 ) * (nvar + 1) + 1] #variance
    avg_summary[3,(2*i-1)] <- res[3,(i - 1 ) * (nvar + 1) + 1] #RMSE
    #averaged slope results
    avg_summary[1,2*i] <- mean(abs(res[1,c(((i - 1 ) * (nvar + 1) + 2):(i *(nvar + 1)))] - coef[-1])) #mean absolute bias
    avg_summary[2,2*i] <- mean(res[2,c(((i - 1 ) * (nvar + 1) + 2):(i *(nvar + 1)))]) #mean variance
    avg_summary[3,2*i] <- sqrt( mean(res[3,c(((i - 1 ) * (nvar + 1) + 2):(i *(nvar + 1)))]^2) ) #mean RMSE
  }
  avg_summary <- round(avg_summary,digits = 4)
  rownames(avg_summary) <- c("bias","var","RMSE")
  colnames(avg_summary) <- rep(c("Intercept","Slope"),q)
  
  OLS <- avg_summary[,1:2]
  KDRE <- avg_summary[,3:4]
  RKDRE <- avg_summary[,5:6]
  SBS <- avg_summary[,7:8]
  LGMMS <- avg_summary[,9:10]
  LGMM <- avg_summary[,11:12]
  YDG <- avg_summary[,13:14]
  
  return(list(summary = list(OLS = OLS,KDRE = KDRE,RKDRE = RKDRE,SBS = SBS,LGMMS = LGMMS,LGMM = LGMM,YDG = YDG), 
              results = res, iterations = iter, estimates = results, perc_errors_YDG = perc_errors))
}

############################################################
#Create graphs and tables
############################################################

#Table 2A
normal.res <- list()
nrep <- 500
for(nvar in c(1,4,9)){
  for(n in c(50,100,500,1000)){
    name <- paste("nvar= ",nvar, ", n= ",n, sep="")
    if (nvar == 1){
      coef <- c(1,-1)
    }
    else if (nvar == 4){
      coef <- c(1,-1,2,-0.5,3)
    }
    else if (nvar == 9){
      coef <- c(1,-1,2,-0.5,3,1,-1,2,-0.5,3)
    }
    print(name)
    normal.res[[name]] <- sim(nrep = nrep, n = n, nvar = nvar, coef = coef, distribution = "normal", bandwidth = "normal")
    print(normal.res[[name]]$summary)
   }
}
save(normal.res, file = "normal.res.RData")

#Table 2B
contam_normal.res <- list()
nrep <- 500
for(nvar in c(1,4,9)){
  for(n in c(50,100,500,1000)){
    name <- paste("nvar= ",nvar, ", n= ",n, sep="")
    if (nvar == 1){
      coef <- c(1,-1)
    }
    else if (nvar == 4){
      coef <- c(1,-1,2,-0.5,3)
    }
    else if (nvar == 9){
      coef <- c(1,-1,2,-0.5,3,1,-1,2,-0.5,3)
    }
    print(name)
    contam_normal.res[[name]] <- sim(nrep = nrep, n = n, nvar = nvar, coef = coef, distribution = "contam_normal", bandwidth = "normal")
    print(contam_normal.res[[name]]$summary)
  }
}
save(contam_normal.res, file = "contam_normal.res.RData")

#Table 2C
heavy_tail.res <- list()
nrep <- 500
for(nvar in c(1,4,9)){
  for(n in c(50,100,500,1000)){
    name <- paste("nvar= ",nvar, ", n= ",n, sep="")
    if (n == 50 | n == 100){
      nrep <- 1000
    } else {
      nrep <- 500
    }
    
    if (nvar == 1){
      coef <- c(1,-1)
    }
    else if (nvar == 4){
      coef <- c(1,-1,2,-0.5,3)
    }
    else if (nvar == 9){
      coef <- c(1,-1,2,-0.5,3,1,-1,2,-0.5,3)
    }
    print(name)
    heavy_tail.res[[name]] <- sim(nrep = nrep, n = n, nvar = nvar, coef = coef, distribution = "heavy_tail", bandwidth = "normal")
    print(heavy_tail.res[[name]]$summary)
  }
}
save(heavy_tail.res, file = "heavy_tail.res.237.RData")

#Table 2D
multi_modal.res <- list()
nrep <- 500
for(nvar in c(1,4,9)){
  for(n in c(50,100,500,1000)){
    name <- paste("nvar= ",nvar, ", n= ",n, sep="")
    if (nvar == 1){
      coef <- c(1,-1)
    }
    else if (nvar == 4){
      coef <- c(1,-1,2,-0.5,3)
    }
    else if (nvar == 9){
      coef <- c(1,-1,2,-0.5,3,1,-1,2,-0.5,3)
    }
    print(name)
    multi_modal.res[[name]] <- sim(nrep = nrep, n = n, nvar = nvar, coef = coef, distribution = "multi_modal", bandwidth = "normal")
    print(multi_modal.res[[name]]$summary)
  }
}
save(multi_modal.res, file = "multi_modal.res.27june.RData")

#Table 2E
uniform.res <- list()
nrep <- 500
for(nvar in c(1,4,9)){
  for(n in c(50,100,500,1000)){
    name <- paste("nvar= ",nvar, ", n= ",n, sep="")
    if (nvar == 1){
      coef <- c(1,-1)
    }
    else if (nvar == 4){
      coef <- c(1,-1,2,-0.5,3)
    }
    else if (nvar == 9){
      coef <- c(1,-1,2,-0.5,3,1,-1,2,-0.5,3)
    }
    print(name)
    uniform.res[[name]] <- sim(nrep = nrep, n = n, nvar = nvar, coef = coef, distribution = "uniform", bandwidth = "normal")
    print(uniform.res[[name]]$summary)
  }
}
save(uniform.res, file = "uniform.res.RData")

#Table 2F
gamma.res <- list()
nrep <- 500
for(nvar in c(1,4,9)){
  for(n in c(50,100,500,1000)){
    name <- paste("nvar= ",nvar, ", n= ",n, sep="")
    if (nvar == 1){
      coef <- c(1,-1)
    }
    else if (nvar == 4){
      coef <- c(1,-1,2,-0.5,3)
    }
    else if (nvar == 9){
      coef <- c(1,-1,2,-0.5,3,1,-1,2,-0.5,3)
    }
    print(name)
    gamma.res[[name]] <- sim(nrep = nrep, n = n, nvar = nvar, coef = coef, distribution = "gamma", bandwidth = "robust2")
    print(gamma.res[[name]]$summary)
  }
}
save(gamma.res, file = "gamma.res.withoutnvar9.n1000.RData")

gamma.res.nvar9.n1000.nrep300 <- sim(nrep = 300, n = 1000, nvar = 9, coef = c(1,-1,2,-0.5,3,1,-1,2,-0.5,3), distribution = "gamma", bandwidth = "robust2")
save(gamma.res.nvar9.n1000.nrep300, file = "gamma.res.nvar9.n1000.nrep300.RData")

#Table 2G
skewed.res <- list()
nrep <- 500
for(nvar in c(1,4,9)){
  for(n in c(50,100,500,1000)){
    name <- paste("nvar= ",nvar, ", n= ",n, sep="")
    if (nvar == 1){
      coef <- c(1,-1)
    }
    else if (nvar == 4){
      coef <- c(1,-1,2,-0.5,3)
    }
    else if (nvar == 9){
      coef <- c(1,-1,2,-0.5,3,1,-1,2,-0.5,3)
    }
    print(name)
    skewed.res[[name]] <- sim(nrep = nrep, n = n, nvar = nvar, coef = coef, distribution = "skewed", bandwidth = "robust2")
    print(skewed.res[[name]]$summary)
  }
}
save(skewed.res, file = "skewed.res.RData")

#Table 2H
log_normal.res <- list()
nrep <- 500
for(nvar in c(1,4,9)){
  for(n in c(50,100,500,1000)){
    name <- paste("nvar= ",nvar, ", n= ",n, sep="")
    if (nvar == 1){
      coef <- c(1,-1)
    }
    else if (nvar == 4){
      coef <- c(1,-1,2,-0.5,3)
    }
    else if (nvar == 9){
      coef <- c(1,-1,2,-0.5,3,1,-1,2,-0.5,3)
    }
    print(name)
    log_normal.res[[name]] <- sim(nrep = nrep, n = n, nvar = nvar, coef = coef, distribution = "log_normal", bandwidth = "robust2")
    print(log_normal.res[[name]]$summary)
  }
}
save(log_normal.res, file = "log_normal.res.RData")





